/*
 * tcpWin.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "tcpWin.h"


/*
 * Plugin variables that may be used by other plugins MUST be declared in
 * the header file as 'extern tcpWinFlow_t *tcpWinFlows;'
 */

tcpWinFlow_t *tcpWinFlows;


/*
 * Static variables are only visible in this file
 */

// window size counts
static gwz_t  gwz;                      // global window size structure

static uint32_t pktTcpCnt;              // Aggregated TCP packet count
static uint32_t winThCntG, winThCntG0;  // Aggregated win threshold count and variable for the last threshold count
static uint8_t  tcpWinStat;             // Aggregated status


/*
 * Static functions prototypes
 */


/*
 * Macros
 */

#define TCPWIN_SPKTMD_PRI_NONE() \
    if (sPktFile) { \
        fputs(/* tcpWinSize     */ SEP_CHR \
              /* tcpWinThPktCnt */ SEP_CHR \
              , sPktFile); \
    }


// Tranalyzer functions

/*
 * This describes the plugin name, version, major and minor version of
 * Tranalyzer required and dependencies
 */
T2_PLUGIN_INIT("tcpWin", "0.9.0", 0, 9);


/*
 * This function is called before processing any packet.
 */
void t2Init() {
    // allocate struct for all flows and initialize to 0
    T2_PLUGIN_STRUCT_NEW(tcpWinFlows);

    // Packet mode
    if (sPktFile) {
        fputs("tcpWinSize"     SEP_CHR // Note the trailing separator SEP_CHR
              "tcpWinThPktCnt" SEP_CHR // Note the trailing separator SEP_CHR
              , sPktFile);
    }
}


/*
 * This function is used to describe the columns output by the plugin
 */
binary_value_t* t2PrintHeader() {
    binary_value_t *bv = NULL;

    BV_APPEND_H8(bv, "tcpWinStat", "TCP window size threshold status");
    BV_APPEND_U8(bv, "tcpWinIpTTL" , "IP TTL");
    BV_APPEND_U32(bv, "tcpInitWinSz", "TCP initial effective window size");
    BV_APPEND_U32(bv, "tcpWinThCnt", "TCP window size threshold count");
    BV_APPEND_FLT(bv, "tcpWinSzThRt", "TCP packet count ratio below window size WINMIN threshold");

    return bv;
}


/*
 * This function is called every time a new flow is created.
 */
void t2OnNewFlow(packet_t *packet UNUSED, unsigned long flowIndex) {
    // Reset the structure for this flow
    tcpWinFlow_t * const tcpWinFlowP = &tcpWinFlows[flowIndex];
    memset(tcpWinFlowP, '\0', sizeof(*tcpWinFlowP));

    const flow_t * const flowP = &flows[flowIndex];
    if (flowP->status & L2_FLOW) return; // Layer 2 flow. No L3/4 pointers, so return

    if (PACKET_IS_IPV6(packet)) {
        const ip6Header_t * const ip6Header = IPV6_HEADER(packet);
        tcpWinFlowP->ttl = ip6Header->ip_ttl;
    } else {
        const ipHeader_t * const ipHeader = IPV4_HEADER(packet);
        tcpWinFlowP->ttl = ipHeader->ip_ttl;
    }

    if (flowP->l4Proto != L3_TCP) return; // process only TCP

    const tcpHeader_t * const tcpHeader = TCP_HEADER(packet);
    tcpWinFlowP->tcpWinInit = ntohs(tcpHeader->window);
}


#if ETH_ACTIVATE > 0
/*
 * This function is called for every packet with a layer 2.
 * If flowIndex is HASHTABLE_ENTRY_NOT_FOUND, this means the packet also
 * has a layer 4 and thus a call to t2OnLayer4() will follow.
 */
void t2OnLayer2(packet_t *packet UNUSED, unsigned long flowIndex) {
    if (flowIndex == HASHTABLE_ENTRY_NOT_FOUND) return;

    // This packet does not have a layer 4.
    // Print the appropriate amount of separators (SEP_CHR) to keep the packet file aligned
    TCPWIN_SPKTMD_PRI_NONE();  // Packet mode
}
#endif // ETH_ACTIVATE > 0


/*
 * This function is called for every packet with a layer 4.
 */
void t2OnLayer4(packet_t *packet, unsigned long flowIndex) {
    const flow_t * const flowP = &flows[flowIndex];

    if (flowP->l4Proto != L3_TCP) {             // process only TCP
        TCPWIN_SPKTMD_PRI_NONE();               // Packet mode
        return;                                 // go back to core
    }

    // only 1. frag packet will be processed
    if (!t2_is_first_fragment(packet)) {
        TCPWIN_SPKTMD_PRI_NONE();               // Packet mode
        return;                                 // go back to core
    }

    tcpWinFlow_t * const tcpWinFlowP = &tcpWinFlows[flowIndex];
    const tcpHeader_t * const tcpHeader = TCP_HEADER(packet);
    const uint32_t tcpWin = ntohs(tcpHeader->window);

    tcpWinFlowP->pktTcpCnt++;                 // count all TCP packets

    if (tcpWin < TCPWIN_THRES) {              // is the window size below the threshold?
        tcpWinFlowP->winThCnt++;              // count the packet / flow
        winThCntG++;                          // increment global packet counter
        tcpWinFlowP->stat |= TCPWIN_STAT_THU; // set the status bit
        tcpWinStat |= tcpWinFlowP->stat;      // Aggregate all packet flags
    }

    // Packet mode
    if (sPktFile) {
        fprintf(sPktFile,
                "%" PRIu32 /* tcpWinSize     */ SEP_CHR
                "%" PRIu32 /* tcpWinThPktCnt */ SEP_CHR
                , tcpWin, tcpWinFlowP->winThCnt);
    }
}


/*
 * This function is called once a flow is terminated.
 * Output all the statistics for the flow here.
 */
void t2OnFlowTerminate(unsigned long flowIndex, outputBuffer_t *buf) {
    const tcpWinFlow_t * const tcpWinFlowP = &tcpWinFlows[flowIndex];

    float f = 0.0;
    if (tcpWinFlowP->pktTcpCnt) {
        f = (float)tcpWinFlowP->winThCnt/(float)tcpWinFlowP->pktTcpCnt;
        pktTcpCnt += tcpWinFlowP->pktTcpCnt;  // Count all TCP packets
    }

    OUTBUF_APPEND_U8(buf, tcpWinFlowP->stat);
    OUTBUF_APPEND_U8(buf, tcpWinFlowP->ttl);
    OUTBUF_APPEND_U32(buf, tcpWinFlowP->tcpWinInit);
    OUTBUF_APPEND_U32(buf, tcpWinFlowP->winThCnt);
    OUTBUF_APPEND_FLT(buf, f);

    const flow_t * const flowP = &flows[flowIndex];
    if (!FLOW_IS_IPV4(flowP)) return;  // IPv4 only

    if (tcpWinFlowP->winThCnt == 0 || tcpWinFlowP->pktTcpCnt < TCPWIN_MINPKTS) return;

    const int wzi = gwz.wzi; // store element count in const local variable, makes the compiler happy
    if (wzi >= TCPWIN_MAXWSCNT) return;  // If array full, stop saving

    int i;
    for (i = 0; i < wzi; i++) {
        if (gwz.wzip[i].IPv4.s_addr == flowP->srcIP.IPv4.s_addr) break; // does IP exist?
    }

    if (f > gwz.wzCnt[i]) {                     // only update if count is greater than the previous one
        gwz.tcpCnt[i] = tcpWinFlowP->pktTcpCnt; // update TCP packet count
        gwz.wzCnt[i] = f;                       // update relative count
        if (i == wzi) {                         // new one?
            gwz.wzip[i] = flowP->srcIP;         // save new IP
#if SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
            gwz.sID[i] = flowP->subnetNrSrc;    // save subnetID from core
#endif // SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
            gwz.wzi++;                          // increment global window size counter
        }
    }
}


/*
 * This function is used to report information regarding the plugin.
 * This will appear in the final report.
 */
void t2PluginReport(FILE *stream) {
    if (winThCntG) {
        T2_FPLOG_AGGR_HEX(stream, plugin_name, tcpWinStat);
        T2_FPLOG_NUMP(stream, plugin_name,
                "Number of TCP winsize packets below threshold " STR(TCPWIN_THRES),
                winThCntG, pktTcpCnt);
        if (gwz.wzi) {
            uint32_t gzCM = gwz.wzCnt[0];
            uint32_t gzi = 0;

            // find IP with largest ipCnt
            for (int i = 0; i < gwz.wzi; i++) {
                if (gwz.wzCnt[i] > gzCM) {
                    gzCM = gwz.wzCnt[i];
                    gzi = i;
                }
            }

            char srcIP[INET_ADDRSTRLEN];
            t2_ipv4_to_str(gwz.wzip[gzi].IPv4, srcIP, sizeof(srcIP));

#if SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
            char *loc, *org;
            SUBNET_LOC(loc, 4, gwz.sID[gzi]);  // <-- get country for IP
            SUBNET_ORG(org, 4, gwz.sID[gzi]);  // <-- get organization for IP
            T2_FPLOG(stream, plugin_name, "IP: %s, country: %s, org: %s", srcIP, loc, org);
#else // SUBNET_ON == 0 || (AGGREGATIONFLAG & SUBNET) != 0
            T2_FPLOG(stream, plugin_name, "IP: %s", srcIP);
#endif // SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
        }
    }
}


/*
 * This function is used to report information regarding the plugin
 * at regular interval or when a USR1 signal is received.
 */
void t2Monitoring(FILE *stream, uint8_t state) {

    switch (state) {

        // Print the name of the variables that will be output
        case T2_MON_PRI_HDR: // hdr output at the beginning
            fputs("winPktThCnt" SEP_CHR, stream); // Note the trailing separator (SEP_CHR)
            return;

        // Print the variables to monitor
        case T2_MON_PRI_VAL: // USR1/2 interrupt machine mode
            // Note the trailing separator (SEP_CHR)
            fprintf(stream, "%" PRIu32 /* winPktThCnt */ SEP_CHR, winThCntG - winThCntG0);
            break;

        // Print a report similar to t2PluginReport()
        case T2_MON_PRI_REPORT: // USR1/2 interrupt report mode
            T2_FPLOG_DIFFNUMP(stream, plugin_name,
                    "Number of TCP winsize packets below threshold" STR(TCPWIN_THRES),
                    winThCntG, numPackets);
            break;

        default:  // Invalid state, do nothing
            return;
    }

#if DIFF_REPORT == 1
    winThCntG0 = winThCntG; // differential mode, save the old value
#endif // DIFF_REPORT == 1
}


/*
 * This function is called once all the packets have been processed.
 * Cleanup all used memory here.
 */
void t2Finalize() {
    free(tcpWinFlows);

    // open tcpWin statistics file
    FILE *fp = t2_fopen_with_suffix(baseFileName, TCPWIN_SUFFIX, "w");
    if (UNLIKELY(!fp)) return;  // return if file could not be opened

    // print header
#if SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
    fprintf(fp, "# IP\tCountry\tOrg\tpktTcpCnt\twinRelThCnt\n");
#else // SUBNET_ON == 0 || (AGGREGATIONFLAG & SUBNET) != 0
    fprintf(fp, "# IP\tpktTcpCnt\twinRelThCnt\n");
#endif // SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0

    char srcIP[INET_ADDRSTRLEN];
    for (int i = 0; i < gwz.wzi; i++) {
        t2_ipv4_to_str(gwz.wzip[i].IPv4, srcIP, sizeof(srcIP)); // convert IP to string
#if SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
        char *loc, *org;
        SUBNET_LOC(loc, 4, gwz.sID[i]); // get country for IP
        SUBNET_ORG(org, 4, gwz.sID[i]); // get organization for IP
        fprintf(fp, "%s\t%s\t%s\t%" PRIu32 "\t%f\n",
                srcIP, loc, org, gwz.tcpCnt[i], gwz.wzCnt[i]);
#else // SUBNET_ON == 0 || (AGGREGATIONFLAG & SUBNET) != 0
        fprintf(fp, "%s\t%" PRIu32 "\t%f\n",
                srcIP, gwz.tcpCnt[i], gwz.wzCnt[i]);
#endif // SUBNET_ON != 0 && (AGGREGATIONFLAG & SUBNET) == 0
    }

    fclose(fp);
}
